#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

from pycutlass import *
import pycutlass
from pycutlass.epilogue import LinearCombination
from pycutlass.test.conv2d_testbed import Conv2dLauncher


if __name__ == "__main__":
    pycutlass.get_memory_pool(2**33, 2**33)
    pycutlass.compiler.nvcc()

    math_inst = MathInstruction(
        instruction_shape=[16, 8, 16],
        element_a=cutlass.float16, element_b=cutlass.float16,
        element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
        math_operation=MathOperation.multiply_add
    )

    A = TensorDescription(
        element=math_inst.element_a, 
        layout=cutlass.TensorNHWC,
        alignment=8)
    B = TensorDescription(
        element=math_inst.element_b, 
        layout=cutlass.TensorNHWC, 
        alignment=8)
    C = TensorDescription(
        element=cutlass.float32,
        layout=cutlass.TensorNHWC, 
        alignment=8)

    tile_description = TileDescription(
        threadblock_shape=[128, 128, 64], stages=4, 
        warp_count=[2, 2, 1],
        math_instruction=math_inst
    )

    epilogue_functor = LinearCombination(cutlass.float32, 4, cutlass.float32, cutlass.float32)

    operation = Conv2dOperation(
        conv_kind=cutlass.conv.Operator.fprop, iterator_algorithm=cutlass.conv.IteratorAlgorithm.optimized,
        arch=80, tile_description=tile_description, A=A, B=B, C=C, 
        element_epilogue=cutlass.float32, stride_support=StrideSupport.Strided,
        epilogue_functor=epilogue_functor,
        swizzling_functor=cutlass.IdentitySwizzle1
    )

    profiler = Conv2dLauncher(operation, verification=False, profiling=True)

    python_runtime = profiler.run(
        problem_size = cutlass.conv.Conv2dProblemSize(
            cutlass.Tensor4DCoord(32, 224, 224, 128),
            cutlass.Tensor4DCoord(128, 3, 3, 128),
            cutlass.Tensor4DCoord(1, 1, 1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.conv.Mode.cross_correlation,
            1, 1
        ), split_k_mode=cutlass.conv.SplitKMode.Serial
    )


    cpp_runtime = profiler.run_cutlass_profiler(
        problem_size = cutlass.conv.Conv2dProblemSize(
            cutlass.Tensor4DCoord(32, 224, 224, 128),
            cutlass.Tensor4DCoord(128, 3, 3, 128),
            cutlass.Tensor4DCoord(1, 1, 1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.conv.Mode.cross_correlation,
            1, 1
        ), split_k_mode=cutlass.conv.SplitKMode.Serial
    )

    print(cpp_runtime / python_runtime)



import pycutlass
from pycutlass import *
from pycutlass.test import *
import unittest

from pycutlass.test.gemm_testbed import test_all_gemm


class GemmF16Sm80(unittest.TestCase):
    def test_SM80_Device_Gemm_f32t_f32n_f32t_tensor_op_bf16_f32_128x128x32_64x64x32(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 32],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float32, layout=cutlass.ColumnMajor,
            alignment=4
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.BatchedIdentitySwizzle

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor,
            direct_store=True
        )

        self.assertTrue(test_all_gemm(operation, "universal"))

    def test_SM80_Device_Gemm_f16n_f16n_f16t_tensor_op_f32_128x128x64_64x64x64(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 64],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_Gemm_f16n_f16n_f32n_tensor_op_f32_128x256x64_64x64x64(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 256, 64],
            stages=3, warp_count=[2, 4, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float32, layout=cutlass.ColumnMajor,
            alignment=4
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_Gemm_f16n_f16n_f32t_tensor_op_f32_256x128x64_64x64x64(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[256, 128, 64],
            stages=3, warp_count=[4, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float32, layout=cutlass.RowMajor,
            alignment=4
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_Gemm_f16n_f16t_f16t_tensor_op_f16_sliced_k_128x64x64_64x64x32(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float16, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 64, 64],
            stages=3, warp_count=[2, 1, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=4
        )

        element_epilogue = cutlass.float16

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_GemmUniversal_f16n_f16t_f32t_tensor_op_f32_64x64x32_32x32x32(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float16, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[64, 64, 32],
            stages=10, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=4
        )

        element_epilogue = cutlass.float16

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_Gemm_f16n_f16t_f32t_tensor_op_f32_256x128x64_64x64x64(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[256, 128, 64],
            stages=3, warp_count=[4, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_test_SM80_Device_Gemm_f16t_f16n_f16t_tensor_op_f16_sliced_k_128x64x64_64x64x32(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 64, 64],
            stages=3, warp_count=[2, 1, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=4
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_Gemm_f16t_f16t_f32n_tensor_op_f32_128x256x64_64x64x64(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 256, 64],
            stages=3, warp_count=[2, 4, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.RowMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_Gemm_f16t_f16t_f32t_tensor_op_f32_128x256x64_64x64x64(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 256, 64],
            stages=3, warp_count=[2, 4, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )
        C = TensorDescription(
            element=cutlass.float32, layout=cutlass.ColumnMajor,
            alignment=4
        )

        element_epilogue = cutlass.float32

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    

if __name__ == '__main__':
    pycutlass.get_memory_pool(2**24, 2**24)
    unittest.main()



import pycutlass
from pycutlass import *
from pycutlass.test import *
import unittest

from pycutlass.test.gemm_grouped_testbed import TestbedGrouped


class GemmGroupedSm80(unittest.TestCase):
    def test_SM80_Device_GemmGrouped_f16n_f16t_f32n_tensor_op_f32_128x128x32_64x64x32(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16], element_a=cutlass.float16,
            element_b=cutlass.float16, element_accumulator=cutlass.float32,
            opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 32],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )

        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )

        C = TensorDescription(
            element=cutlass.float32, layout=cutlass.ColumnMajor,
            alignment=4
        )

        element_epilogue = cutlass.float32
        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        swizzling_functor = cutlass.BatchedIdentitySwizzle

        for precompute_mode in [SchedulerMode.Device, SchedulerMode.Host]:
            operation = GemmOperationGrouped(
                80,
                tile_description, A, B, C,
                epilogue_functor, swizzling_functor,
                precompute_mode=precompute_mode
            )

            testbed = TestbedGrouped(operation=operation)

            self.assertTrue(testbed.run(24))
    
    def test_SM80_Device_GemmGrouped_f64t_f64t_f64n_tensor_op_f64_64x64x16_32x32x16(self):
        math_inst = MathInstruction(
            instruction_shape=[8, 8, 4], element_a=cutlass.float64,
            element_b=cutlass.float64, element_accumulator=cutlass.float64,
            opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[64, 64, 16],
            stages=4, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float64, layout=cutlass.RowMajor,
            alignment=1
        )

        B = TensorDescription(
            element=cutlass.float64, layout=cutlass.RowMajor,
            alignment=1
        )

        C = TensorDescription(
            element=cutlass.float64, layout=cutlass.ColumnMajor,
            alignment=1
        )

        element_epilogue = cutlass.float64
        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        swizzling_functor = cutlass.BatchedIdentitySwizzle

        for precompute_mode in [SchedulerMode.Device, SchedulerMode.Host]:
            operation = GemmOperationGrouped(
                80,
                tile_description, A, B, C,
                epilogue_functor, swizzling_functor,
                precompute_mode=precompute_mode
            )

            testbed = TestbedGrouped(operation=operation)

            self.assertTrue(testbed.run(24))
    
    def test_SM80_Device_GemmGrouped_f32t_f32t_f32t_simt_f32_128x64x8_64x32x1(self):
        math_inst = MathInstruction(
            instruction_shape=[1, 1, 1], element_a=cutlass.float32,
            element_b=cutlass.float32, element_accumulator=cutlass.float32,
            opcode_class=cutlass.OpClass.Simt,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 64, 8],
            stages=4, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float32, layout=cutlass.RowMajor,
            alignment=1
        )

        B = TensorDescription(
            element=cutlass.float32, layout=cutlass.RowMajor,
            alignment=1
        )

        C = TensorDescription(
            element=cutlass.float32, layout=cutlass.RowMajor,
            alignment=1
        )

        element_epilogue = cutlass.float32
        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        swizzling_functor = cutlass.BatchedIdentitySwizzle

        for precompute_mode in [SchedulerMode.Device, SchedulerMode.Host]:
            operation = GemmOperationGrouped(
                80,
                tile_description, A, B, C,
                epilogue_functor, swizzling_functor,
                precompute_mode=precompute_mode
            )

            testbed = TestbedGrouped(operation=operation)

            self.assertTrue(testbed.run(27))
    
    def test_SM80_Device_GemmGrouped_f16n_f16t_f32n_tensor_op_f32_128x128x32_64x64x32_cache(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16], element_a=cutlass.float16,
            element_b=cutlass.float16, element_accumulator=cutlass.float32,
            opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 32],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )

        B = TensorDescription(
            element=cutlass.float16, layout=cutlass.ColumnMajor,
            alignment=8
        )

        C = TensorDescription(
            element=cutlass.float32, layout=cutlass.ColumnMajor,
            alignment=4
        )

        element_epilogue = cutlass.float32
        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        swizzling_functor = cutlass.BatchedIdentitySwizzle

        for precompute_mode in [SchedulerMode.Device, SchedulerMode.Host]:
            operation = GemmOperationGrouped(
                80,
                tile_description, A, B, C,
                epilogue_functor, swizzling_functor,
                precompute_mode=precompute_mode
            )

            testbed = TestbedGrouped(operation=operation)

            self.assertTrue(testbed.run(5))


if __name__ == '__main__':
    pycutlass.get_memory_pool(2**26, 2**26)
    unittest.main()



import pycutlass
from pycutlass import *
from pycutlass.epilogue import LinearCombinationClamp
from pycutlass.test import *
import unittest

from pycutlass.test.gemm_testbed import test_all_gemm

class GemmS8TensorOpF32Sm80(unittest.TestCase):
    def test_SM80_Device_Gemm_s8t_s8n_s8t_tensor_op_s32_64x64x64_32x32x64(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 32],
            element_a=cutlass.int8, element_b=cutlass.int8,
            element_accumulator=cutlass.int32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add_saturate
        )

        tile_description = TileDescription(
            threadblock_shape=[64, 64, 64],
            stages=6, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.int8, layout=cutlass.ColumnMajorInterleaved32,
            alignment=16
        )
        B = TensorDescription(
            element=cutlass.int8, layout=cutlass.RowMajorInterleaved32,
            alignment=16
        )
        C = TensorDescription(
            element=cutlass.int8, layout=cutlass.ColumnMajorInterleaved32,
            alignment=8
        )

        epilogue_functor = FastLinearCombinationClamp(
            C.element, C.alignment
        )
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C,
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "interleaved"))
    
    def test_SM80_Device_Gemm_s8t_s8n_s8t_tensor_op_s32_256x128x128_64x64x128(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 32],
            element_a=cutlass.int8, element_b=cutlass.int8,
            element_accumulator=cutlass.int32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 128],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.int8, layout=cutlass.RowMajor,
            alignment=16
        )
        B = TensorDescription(
            element=cutlass.int8, layout=cutlass.ColumnMajor,
            alignment=16
        )
        C = TensorDescription(
            element=cutlass.int8, layout=cutlass.RowMajor,
            alignment=16
        )

        epilogue_functor = FastLinearCombinationClamp(
            C.element, C.alignment
        )
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C,
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "multistage"))
    
    def test_SM80_Device_Gemm_s8t_s8n_s8n_tensor_op_s32_128x128x128_64x64x128(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 32],
            element_a=cutlass.int8, element_b=cutlass.int8,
            element_accumulator=cutlass.int32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 128],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.int8, layout=cutlass.RowMajor,
            alignment=16
        )
        B = TensorDescription(
            element=cutlass.int8, layout=cutlass.ColumnMajor,
            alignment=16
        )
        C = TensorDescription(
            element=cutlass.int8, layout=cutlass.ColumnMajor,
            alignment=16
        )

        epilogue_functor = FastLinearCombinationClamp(
            C.element, C.alignment
        )
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C,
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "multistage"))
    
    def test_SM80_Device_Gemm_s8t_s8n_s32n_tensor_op_s32_128x128x128_64x64x128(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 32],
            element_a=cutlass.int8, element_b=cutlass.int8,
            element_accumulator=cutlass.int32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 128],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.int8, layout=cutlass.RowMajor,
            alignment=16
        )
        B = TensorDescription(
            element=cutlass.int8, layout=cutlass.ColumnMajor,
            alignment=16
        )
        C = TensorDescription(
            element=cutlass.int32, layout=cutlass.ColumnMajor,
            alignment=4
        )

        element_epilogue = cutlass.int32

        epilogue_functor = LinearCombinationClamp(
            C.element, C.alignment, math_inst.element_accumulator, 
            element_epilogue
        )
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "multistage"))
    
    def test_SM80_Device_Gemm_s8t_s8n_s32t_tensor_op_s32_128x128x128_64x64x128(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 32],
            element_a=cutlass.int8, element_b=cutlass.int8,
            element_accumulator=cutlass.int32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 128],
            stages=3, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        A = TensorDescription(
            element=cutlass.int8, layout=cutlass.RowMajor,
            alignment=16
        )
        B = TensorDescription(
            element=cutlass.int8, layout=cutlass.ColumnMajor,
            alignment=16
        )
        C = TensorDescription(
            element=cutlass.int32, layout=cutlass.RowMajor,
            alignment=4
        )

        element_epilogue = cutlass.int32

        epilogue_functor = LinearCombinationClamp(
            C.element, C.alignment, math_inst.element_accumulator, 
            element_epilogue
        )
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C,
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "multistage"))
    



if __name__ == '__main__':
    pycutlass.get_memory_pool(2**24, 2**24)
    unittest.main()



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'PyCutlass'
copyright = '2022, Zhaodong Chen; Andrew Kerr; Haicheng Wu; Szymon Migacz; Graham Markall'
author = 'Zhaodong Chen; Andrew Kerr; Haicheng Wu; Szymon Migacz; Graham Markall'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'enum_tools.autoenum',
    'sphinx.ext.autosummary',
    'm2r2'
]

source_suffix = [".rst", ".md"]

autosummary_generate = True
autosummary_imported_members = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'bizstyle'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

import ctypes
from cuda import cuda

################################################################################
#
# Launch configuration
#
################################################################################


class LaunchConfiguration:
    def __init__(self, grid=[1, 1, 1], block=[1, 1, 1], smem=0):
        self.grid = grid
        self.block = block
        self.shared_memory_capacity = smem


################################################################################
#
# Base class for an executable operation
#
# ##############################################################################

class ExecutableOperation:
    '''
    '''

    def __init__(self, operation):
        self.operation = operation
        self.module = None
        self.kernel = None

    #
    def name(self):
        return self.operation.procedural_name()

    #
    def emit(self):
        return ''

    #
    def can_implement(self, configuration, arguments):
        raise NotImplementedError()

    #
    def get_host_workspace_size(self, arguments):
        raise NotImplementedError()

    #
    def get_device_workspace_size(self, arguments):
        raise NotImplementedError()

    #
    def plan(self, arguments):
        raise NotImplementedError()

    #
    def initialize(self, host_workspace, device_workspace, launch_config, arguments, stream=cuda.CUstream(0)):
        raise NotImplementedError()

    #
    def run(self, host_workspace, device_workspace, launch_config, stream=cuda.CUstream(0)):

        cArg = (ctypes.c_char * len(host_workspace)
                ).from_buffer(host_workspace)
        packed = (ctypes.c_void_p * 1)()
        packed[0] = ctypes.addressof(cArg)

        err, = cuda.cuLaunchKernel(
            self.kernel,
            launch_config.grid[0], launch_config.grid[1], launch_config.grid[2],
            launch_config.block[0], launch_config.block[1], launch_config.block[2],
            launch_config.shared_memory_capacity,
            stream,
            packed,
            0)

        return err



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

# System imports
import struct
import io
import ctypes

# CUDA Python import
from cuda import cuda
from cuda import nvrtc

# CUTLASS imports
from library import *
from gemm_operation import EmitGemmUniversalInstance

#################################################################################################
#
# CUTLASS Py Runtime Components
#
#################################################################################################

#
def MaxAlignment(fmt):
  align = 1
  for x in fmt:
    align = max(align, struct.calcsize(x))
  return align
  
#
def AlignedOffset(offset, align):
  remainder = (offset % align)
  if remainder:
    offset += (align - remainder)
  return offset

#
def PackInteger(host_workspace, offset, value):
  fmt = "i"
  padding = AlignedOffset(offset, 4)  
  struct.pack_into(fmt, host_workspace, offset, value)
  return padding + struct.calcsize(fmt)

#
def PackDevicePointer(host_workspace, offset, value):
  fmt = "P"
  offset = AlignedOffset(offset, 8)
  struct.pack_into(fmt, host_workspace, offset, value)
  return offset + struct.calcsize(fmt)
  
#
def ceil_div(a, b):
  return -(a // -b)

#################################################################################################

#
class PitchLinearCoord:
  def __init__(self, contiguous, strided):
    self.contiguous = contiguous
    self.strided = strided

#
class GemmCoord:
  def __init__(self, m = 1, n = 1, k = 1):
    self.m = m
    self.n = n
    self.k = k
    self.fmt = "iii"

  #
  def ceil_div(self, rhs):
    return GemmCoord(ceil_div(self.m, rhs.m), ceil_div(self.n, rhs.n), ceil_div(self.k, rhs.k))

  #
  def size(self):
    return struct.calcsize(self.fmt)

  #
  def alignment(self):
    return MaxAlignment(self.fmt)

  #
  def pack_into(self, host_workspace, offset):
    
    offset = AlignedOffset(offset, 4)
  
    struct.pack_into(
      self.fmt, 
      host_workspace, 
      offset, 
      self.m, self.n, self.k)

    return offset + self.size()

#
class TensorRef:
  def __init__(self, pointer = None, layout = 0):
    self.pointer = pointer
    self.layout = layout

  def __str__(self):
    return "(%x, %d)" % (self.pointer._ptr, self.layout)

#################################################################################################

#
class PredicatedTileAccessIteratorDesc:
  '''
  '''
  
  def __init__(
      self, 
      element_size_bits, 
      advance_rank, 
      threadblock_shape, 
      threadmap_iterations, 
      threadmap_delta):

    self.element_size_bits = element_size_bits
    self.advance_rank = advance_rank
    self.threadblock_shape = threadblock_shape
    self.threadmap_iterations = threadmap_iterations
    self.threadmap_delta = threadmap_delta

#
class PredicatedTileAccessIteratorParams:
  '''
  '''
  #
  def __init__(self, desc, label):
    self.desc = desc
    self.label = label
    self.fmt = "qqqq"
  #
  def size(self):
    return struct.calcsize(self.fmt)

  #
  def alignment(self):
    return MaxAlignment(self.fmt)

  #
  def initialize(self, host_workspace, offset, stride):

    offset = AlignedOffset(offset, self.alignment())

    inc_strided = stride *                            \
                  self.desc.threadmap_delta.strided * \
                  self.desc.element_size_bits // 8

    if self.desc.advance_rank:
      inc_advance = self.desc.threadblock_shape.strided * \
                          stride *                        \
                          self.desc.element_size_bits // 8
    else:
      inc_advance = self.desc.threadblock_shape.contiguous * \
                          self.desc.element_size_bits // 8

    inc_next = inc_advance - (self.desc.threadmap_iterations.strided - 1) * \
                      self.desc.threadmap_delta.strided *                   \
                      stride *                                              \
                      self.desc.element_size_bits // 8

    struct.pack_into(
      self.fmt, 
      host_workspace, 
      offset, 
      stride, inc_strided, inc_next, inc_advance)

    return offset + self.size()
  #

#################################################################################################

#
class EpilogueTileDesc:
  '''
  '''
  def __init__(self, column, row, group, cluster, tile):
    self.column = column
    self.row = row
    self.group = group
    self.cluster = cluster
    self.tile = tile

#
class EpilogueThreadMap:
  '''
  '''
  def __init__(self, threads, elements_per_access, element_size_bits, shape, iterations, delta, count):
    self.threads = threads
    self.elements_per_access = elements_per_access
    self.element_size_bits = element_size_bits
    self.shape = shape
    self.iterations = iterations
    self.delta = delta
    self.count = count
    pass

#
class EpilogueTileIteratorParams:
  '''
  '''
  #
  def __init__(self, desc, label):
    self.desc = desc
    self.label = label
    self.fmt = "qqqqqqqq"

  #
  def size(self):
    return struct.calcsize(self.fmt)

  #
  def alignment(self):
    return MaxAlignment(self.fmt)

  #
  def initialize(self, host_workspace, offset, stride):

    stride = stride * self.desc.element_size_bits // 8

    offset = AlignedOffset(offset, self.alignment())

    increment_row = stride * self.desc.delta.row

    increment_group = stride * self.desc.delta.group \
      - stride * self.desc.delta.row * (self.desc.iterations.row - 1)

    increment_cluster = stride * self.desc.delta.cluster \
      - stride * self.desc.delta.group * (self.desc.iterations.group - 1) \
      - stride * self.desc.delta.row * (self.desc.iterations.row - 1)
      
    advance_row = stride * self.desc.shape.row

    advance_group = stride *                   \
      (self.desc.shape.group - 1) * \
      self.desc.shape.row *         \
      self.desc.count.row

    advance_cluster = stride * \
      self.desc.count.group * \
      self.desc.shape.group * \
      self.desc.count.row   * \
      self.desc.shape.row

    advance_tile = stride * \
      self.desc.shape.group * \
      self.desc.shape.row   * \
      self.desc.shape.cluster * \
      self.desc.shape.tile

    struct.pack_into(
      self.fmt,                                           \
      host_workspace,                                     \
      offset,                                             \
      stride,                                             \
      increment_row, increment_group, increment_cluster,  \
      advance_row, advance_group, advance_cluster, advance_tile)

    return offset + self.size()
#

#################################################################################################
#
# Launch configuration
#
#################################################################################################

class LaunchConfiguration:
  def __init__(self, grid = [1,1,1], block = [1,1,1], smem = 0):
    self.grid = grid
    self.block = block
    self.shared_memory_capacity = smem

#################################################################################################
#
# Functors
#
#################################################################################################

#
class Functor:
  def __init__(self):
    self.decl = ''
    self.definition = ''
    self.fmt = ''
    self.identifier = ''

  #
  def emit_declaration(self):
    return self.decl

  #
  def emit_definition(self):
    return self.definition

  # 
  def size(self):
    '''
    Size of the packed Params structure
    '''
    return struct.calcsize(self.fmt)

  #
  def alignment(self):
    return MaxAlignment(self.fmt)

  # 
  def initialize(self, host_workspace, offset, arguments):
    return offset + self.size()

#################################################################################################

#
class LinearCombinationFunctorArguments:
  def __init__(self, alpha = 1.0, beta = 0.0):
    self.alpha = alpha
    self.beta = beta
    self.alpha_ptr = 0
    self.beta_ptr = 0

#
class LinearCombinationFunctor(Functor):
  def __init__(self):
    super().__init__()

    self.decl = """
    cutlass::epilogue::thread::LinearCombination<
      float,
      1,
      float,
      float
    >"""
    self.identifier = 'linear_combination'
    self.fmt = "ffPP"

  # 
  def size(self):
    '''
    Size of the packed Params structure
    '''
    return struct.calcsize(self.fmt)

  #
  def alignment(self):
    return MaxAlignment(self.fmt)

  # 
  def initialize(self, host_workspace, offset, arguments):

    offset = AlignedOffset(offset, self.alignment())

    struct.pack_into(
      self.fmt, 
      host_workspace, offset, 
      arguments.alpha, arguments.beta, arguments.alpha_ptr, arguments.beta_ptr)

    return offset + self.size()

#################################################################################################
#
# Base class for an executable operation
#
#################################################################################################

#
class ExecutableOperation:
  '''
  '''
  def __init__(self, operation):
    self.operation = operation
    self.module = None
    self.kernel = None

  #
  def name(self):
    return self.operation.procedural_name()

  #
  def emit(self):
    return ''

  #
  def can_implement(self, configuration, arguments):
    return False

  #
  def get_host_workspace_size(self, arguments):
    return 0

  #
  def get_device_workspace_size(self, arguments):
    return 0

  #
  def plan(self, arguments):
    return LaunchConfiguration()

  #
  def initialize(self, host_workspace, device_workspace, launch_config, arguments, stream = cuda.CUstream(0)):
    raise NotImplementedError()

  #
  def run(self, host_workspace, device_workspace, launch_config, stream = cuda.CUstream(0)):

    cArg = (ctypes.c_char * len(host_workspace)).from_buffer(host_workspace)
    packed = (ctypes.c_void_p * 1)()
    packed[0] = ctypes.addressof(cArg)

    err, = cuda.cuLaunchKernel(
      self.kernel, 
      launch_config.grid[0], launch_config.grid[1], launch_config.grid[2], 
      launch_config.block[0], launch_config.block[1], launch_config.block[2], 
      launch_config.shared_memory_capacity, 
      stream, 
      packed, 
      0)

    return err

#################################################################################################


#
class GemmArguments:
  '''
  '''
  def __init__(self):
    self.problem_size = GemmCoord(0, 0, 0)
    self.A = TensorRef()
    self.B = TensorRef()
    self.C = TensorRef()
    self.D = TensorRef()
    self.output_op = LinearCombinationFunctorArguments()

#
class ThreadblockSwizzle:
  def __init__(self, threadblock_shape, log_threadblock_cohort = 0):
    self.threadblock_shape = threadblock_shape
    self.log_threadblock_cohort = log_threadblock_cohort

  def grid_tiled_shape(self, problem_size):
    return GemmCoord(
      ceil_div(problem_size.m, self.threadblock_shape.m), 
      ceil_div(problem_size.n, self.threadblock_shape.n), 
      1)

#
class Gemm(ExecutableOperation):
  '''
  GEMM manages the CUTLASS runtime components
  '''
  #
  def __init__(self, operation):
    super().__init__(operation)

    self.emitter = EmitGemmUniversalInstance('_type')
    self.threadblock_swizzle = ThreadblockSwizzle(GemmCoord(128, 128, 8))

    self.threads = 256
    self.shared_memory_capacity = (32 << 10)

    self.params_A = PredicatedTileAccessIteratorParams(
        PredicatedTileAccessIteratorDesc(
          32, 
          1, 
          PitchLinearCoord(128, 8), 
          PitchLinearCoord(1, 4), 
          PitchLinearCoord(1, 2)), 'A')

    self.params_B = PredicatedTileAccessIteratorParams(
        PredicatedTileAccessIteratorDesc(
          32, 
          1, 
          PitchLinearCoord(128, 8), 
          PitchLinearCoord(1, 4), 
          PitchLinearCoord(1, 2)), 'B')

    self.params_C = EpilogueTileIteratorParams(
      EpilogueThreadMap(
        256, 
        1, 
        32,
        EpilogueTileDesc(128, 1, 4, 4, 1), 
        EpilogueTileDesc(4, 1, 2, 1, 1), 
        EpilogueTileDesc(32, 1, 8, 1, 1), 
        EpilogueTileDesc(1, 4, 2, 1, 8)), 'C')

    self.params_D = EpilogueTileIteratorParams(
      EpilogueThreadMap(
        256, 
        1,
        32,
        EpilogueTileDesc(128, 1, 4, 4, 1), 
        EpilogueTileDesc(4, 1, 2, 1, 1), 
        EpilogueTileDesc(32, 1, 8, 1, 1), 
        EpilogueTileDesc(1, 4, 2, 1, 8)), 'D')

    self.output_op = LinearCombinationFunctor()

  #
  def emit(self):
    return self.emitter.emit(self.operation)

  #
  def can_implement(self, configuration, arguments):
    pass

  #
  def get_host_workspace_size(self, arguments):
    return 336

  #
  def get_device_workspace_size(self, arguments):
    return 0

  #
  def plan(self, arguments):
    grid = self.threadblock_swizzle.grid_tiled_shape(arguments.problem_size)
    return LaunchConfiguration([grid.m, grid.n, grid.k], [self.threads, 1, 1], self.shared_memory_capacity)

  #
  def initialize(self, host_workspace, device_workspace, launch_config, arguments, stream = cuda.CUstream(0)):
    
    offset = 0

    # Compute intermediate results
    swizzle_log_tile = 0
    gemm_mode = 0
    batch_count = 1
    gemm_k_size = arguments.problem_size.k

    # Pack into the host workspace buffer
    offset = arguments.problem_size.pack_into(host_workspace, offset)

    grid_tiled_shape = self.threadblock_swizzle.grid_tiled_shape(arguments.problem_size)
    offset = grid_tiled_shape.pack_into(host_workspace, offset)

    offset = PackInteger(host_workspace, offset, swizzle_log_tile)

    offset = self.params_A.initialize(host_workspace, offset, arguments.A.layout)
    offset = self.params_B.initialize(host_workspace, offset, arguments.B.layout)
    offset = self.params_C.initialize(host_workspace, offset, arguments.C.layout)
    offset = self.params_D.initialize(host_workspace, offset, arguments.D.layout)

    offset = self.output_op.initialize(host_workspace, offset, arguments.output_op)

    offset = PackInteger(host_workspace, offset, gemm_mode)
    offset = PackInteger(host_workspace, offset, batch_count)
    offset = PackInteger(host_workspace, offset, gemm_k_size)
    offset = PackDevicePointer(host_workspace, offset, int(arguments.A.pointer))
    offset = PackDevicePointer(host_workspace, offset, int(arguments.B.pointer))
    offset = PackDevicePointer(host_workspace, offset, int(arguments.C.pointer))
    offset = PackDevicePointer(host_workspace, offset, int(arguments.D.pointer))   

    return offset


#################################################################################################
#
# Module represents a compilation unit 
#
#################################################################################################

#
class CompilationOptions:
  '''
  Compilation options.
  '''

  #
  def __init__(self, architectures = [80], include_paths = []):
    self.includes = []
    self.include_paths = include_paths
    self.flags = ['-std=c++11', '-default-device']
    self.architectures = architectures

  #
  def get(self):
    options = []

    for flag in self.flags:
      options.append(bytes(str.encode(flag)))

    for incl in self.include_paths:
      options.append(bytes(str.encode('--include-path=%s' % incl)))

    arch_list = "-arch="
    for idx, arch in enumerate(self.architectures):
      if idx:
        arch_list += ","
      arch_list += "sm_%d" % arch

    options.append(bytes(str.encode(arch_list)))

    return options

IncludeTemplate = r'''#include "${include}"
'''

KernelTemplate = r'''
extern "C"
__global__ void
${operation_name}(${operation_name}${operation_suffix}::Params params) {

  // Dynamic shared memory base pointer
  extern __shared__ int SharedStorageBase[];

  // Declare pointer to dynamic shared memory.
  ${operation_name}${operation_suffix}::SharedStorage *shared_storage =
      reinterpret_cast<${operation_name}${operation_suffix}::SharedStorage *>(SharedStorageBase);

  ${operation_name}${operation_suffix} op;

  op(params, *shared_storage);
}

'''

#
class Module:
  def __init__(self, name, operations, compilation_options):
    self.name = name
    self.operations = operations
    self.module = None
    self.log = None
    self.cubin_image = None
    self.source_buffer = ''

    #
    # Emit source
    #
    self.emit_()

    #
    # Compile
    #
    self.compile_(compilation_options)

    #
    # Load module
    #
    self.load_()
    
    # Done
    return

  # Emit a source buffer
  def emit_(self):

    # 1. Includes
    includes = []
    for operation in self.operations:
      for incl in operation.emitter.includes:
        if incl not in includes:
          includes.append(incl)

    for incl in includes:
      self.source_buffer += SubstituteTemplate(IncludeTemplate, { 'include': incl} )

    # 2. Operations
    for operation in self.operations:
      self.source_buffer += operation.emit()
      values = {
        'operation_name': operation.name(),
        'operation_suffix': operation.emitter.operation_suffix
      }
      self.source_buffer += SubstituteTemplate(KernelTemplate, values)

    # Done
    return

  # Compile with NVRTC
  def compile_(self, compilation_options):

    err, program = nvrtc.nvrtcCreateProgram(
      str.encode(self.source_buffer), 
      bytes(str.encode(self.name)), 
      0, [], [])

    if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
        raise RuntimeError('NVRTC Error: {}'.format(err))

    # Compile program
    options = compilation_options.get()

    err, = nvrtc.nvrtcCompileProgram(program, len(options), options)
    if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:

      error_string = 'NVRTC Error: {}\n'.format(err)

      # Get log from compilation
      err, logSize = nvrtc.nvrtcGetProgramLogSize(program)
      if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
          raise RuntimeError('NVRTC Error: {}'.format(err))
      
      self.log = b' ' * logSize
      err, = nvrtc.nvrtcGetProgramLog(program, self.log)
      if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
          raise RuntimeError('NVRTC Error: {}'.format(err))
      
      raise RuntimeError(error_string + self.log.decode() + self.source_buffer)

    # Get data from compilation
    err, dataSize = nvrtc.nvrtcGetCUBINSize(program)
    if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
        raise RuntimeError('NVRTC Error: {}'.format(err))
    
    self.cubin_image = b' ' * dataSize
    err, = nvrtc.nvrtcGetCUBIN(program, self.cubin_image)
    if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
        raise RuntimeError('NVRTC Error: {}'.format(err))

    return
    
  #
  def load_(self):

    # Load data as module data
    err, self.module = cuda.cuModuleLoadData(self.cubin_image)
    if err != cuda.CUresult.CUDA_SUCCESS:
        raise RuntimeError('Cuda Error: {}'.format(err))
    
    # Get functions
    for operation in self.operations:
      err, operation.kernel = cuda.cuModuleGetFunction(
        self.module, 
        bytes(str.encode(operation.name())))

      if err != cuda.CUresult.CUDA_SUCCESS:
          raise RuntimeError('Cuda Error: {}'.format(err))

      operation.module = self

    return


#################################################################################################
#
# Manifest represents an 'owner' for modules and operations
#
#################################################################################################

#
class Manifest:

  #
  def __init__(self):
    self.operations = {}
    self.modules = []
    pass

  #
  def append_module(self, module):
    '''
    Appends a module and takes ownership of operations used to construct it.
    '''
    
    self.modules.append(module)

    for operation in module.operations:
      self.operations[operation.name()] = operation


#################################################################################################



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

import pycutlass
from pycutlass.test.gemm_testbed import getTensorRef, getTensorView, transpose
from pycutlass import *
import numpy as np
import cutlass
from bfloat16 import bfloat16


class TestbedGrouped:
    def __init__(self, operation: GemmOperationGrouped, seed: int = 2080) -> None:

        pycutlass.compiler.add_module([operation])

        self.seed = seed

        self.operation = operation

        element_size = DataTypeSize[operation.A.element]

        self.dtype_A = self.numpy_type(operation.A.element)
        self.dtype_B = self.numpy_type(operation.B.element)
        self.dtype_C = self.numpy_type(operation.C.element)
        self.dtype_D = self.numpy_type(operation.C.element)

        if element_size == 1:
            self.scope_max = 1
            self.scope_min = 0
        elif element_size <= 8:
            self.scope_max = 1
            self.scope_min = -1
        elif element_size == 16:
            self.scope_max = 4
            self.scope_min = -4
        else:
            self.scope_max = 8
            self.scope_min = -8

        #: compute type
        self.compute_type = operation.epilogue_functor.element_epilogue

        self.accumulator_type = operation.tile_description.math_instruction.element_accumulator

    @staticmethod
    def numpy_type(type):
        if type == cutlass.float64:
            return np.float64
        elif type == cutlass.float32:
            return np.float32
        elif type == cutlass.float16:
            return np.float16
        elif type == cutlass.bfloat16:
            return bfloat16
        elif type == cutlass.int32:
            return np.int32
        elif type == cutlass.int8:
            return np.int8
        else:
            raise ValueError("unsupported type: %s" % ShortDataTypeNames[type])

    def uniform_init(self, size, dtype):
        if dtype in [np.float32, np.float16, bfloat16, np.float64]:
            return np.ceil(
                np.random.uniform(
                    low=self.scope_min - 0.5, high=self.scope_max - 0.5,
                    size=size).astype(dtype)
            )
        else:
            return np.random.uniform(
                low=self.scope_min - 1, high=self.scope_max + 1,
                size=size).astype(dtype)

    def print_problem_size(self, p):
        problem_size = "problem: %d, %d, %d\n" % (p.m(), p.n(), p.k())
        print(problem_size)

    def run(self, problem_count: int, alpha: float = 1.0, beta: float = 0.0) -> bool:

        assert get_allocated_size(
        ) == 0, "%d byte of pool memory is not released in previous run" % get_allocated_size()

        # initialize
        np.random.seed(self.seed)

        # generate the problem sizes
        problem_sizes = []
        tensor_As = []
        tensor_Bs = []
        tensor_Cs = []
        tensor_Ds = []
        tensor_D_refs = []

        for i in range(problem_count):
            if self.dtype_A == np.int8:
                if i == 0:
                    problem_size = cutlass.gemm.GemmCoord(48, 16, 32)
                else:
                    problem_size = cutlass.gemm.GemmCoord(
                        16 * np.random.randint(0, 64) + 48,
                        16 * np.random.randint(0, 64) + 48,
                        16 * np.random.randint(0, 64) + 48
                    )
            else:
                if i == 0:
                    problem_size = cutlass.gemm.GemmCoord(48, 16, 8)
                else:
                    problem_size = cutlass.gemm.GemmCoord(
                        8 * np.random.randint(0, 64) + 24,
                        8 * np.random.randint(0, 64) + 24,
                        8 * np.random.randint(0, 64) + 24
                    )

            tensor_As.append(
                self.uniform_init(
                    size=(problem_size.m() * problem_size.k(),),
                    dtype=self.dtype_A)
            )
            tensor_Bs.append(
                self.uniform_init(
                    size=(problem_size.n() * problem_size.k(),),
                    dtype=self.dtype_B)
            )
            tensor_Cs.append(
                self.uniform_init(
                    size=(problem_size.m() * problem_size.n(),),
                    dtype=self.dtype_C)
            )

            tensor_Ds.append(
                np.zeros(
                    shape=(problem_size.m() * problem_size.n(),),
                    dtype=self.dtype_D
                )
            )

            tensor_D_refs.append(
                np.ones(
                    shape=(problem_size.m() * problem_size.n(),),
                    dtype=self.dtype_D
                )
            )

            problem_sizes.append(problem_size)

        arguments = GemmGroupedArguments(
            operation=self.operation, problem_sizes=problem_sizes,
            A=tensor_As, B=tensor_Bs, C=tensor_Cs, D=tensor_Ds,
            output_op=self.operation.epilogue_type(alpha, beta)
        )

        self.operation.run(arguments)

        arguments.sync()

        #
        # Reference check - TODO: support caching results
        #
        alpha = self.compute_type(alpha).value()
        beta = self.compute_type(beta).value()
        init_acc = self.accumulator_type(0).value()

        for idx, problem_size in enumerate(problem_sizes):
            if self.operation.switched:
                tensor_ref_A = getTensorRef(
                    tensor_As[idx], problem_size, "a", transpose(self.operation.B.layout))
                tensor_ref_B = getTensorRef(
                    tensor_Bs[idx], problem_size, "b", transpose(self.operation.A.layout))
                tensor_ref_C = getTensorRef(
                    tensor_Cs[idx], problem_size, "c", transpose(self.operation.C.layout))
                tensor_ref_D_ref = getTensorRef(
                    tensor_D_refs[idx], problem_size, "d", transpose(self.operation.C.layout))
            else:
                tensor_ref_A = getTensorRef(
                    tensor_As[idx], problem_size, "a", self.operation.A.layout)
                tensor_ref_B = getTensorRef(
                    tensor_Bs[idx], problem_size, "b", self.operation.B.layout)
                tensor_ref_C = getTensorRef(
                    tensor_Cs[idx], problem_size, "c", self.operation.C.layout)
                tensor_ref_D_ref = getTensorRef(
                    tensor_D_refs[idx], problem_size, "d", self.operation.C.layout)

            tensor_view_D_ref = getTensorView(
                tensor_D_refs[idx], problem_size, "d", self.operation.C.layout)

            cutlass.test.gemm.host.gemm(problem_size, alpha, tensor_ref_A,
                                        tensor_ref_B, beta, tensor_ref_C, tensor_ref_D_ref, init_acc)

            tensor_view_D = getTensorView(
                tensor_Ds[idx], problem_size, "d", self.operation.C.layout)

            passed = cutlass.test.gemm.host.equals(
                tensor_view_D, tensor_view_D_ref)

            try:
                assert passed
            except AssertionError:
                self.print_problem_size(problem_size)

        del arguments

        assert get_allocated_size(
        ) == 0, "%d byte of pool memory is not released after current run" % get_allocated_size()

        return passed



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

from time import sleep
import pycutlass
from pycutlass import *
import cutlass
from cuda import cudart
from cuda import cuda
from bfloat16 import bfloat16
from .profiler import GpuTimer
import subprocess


def transpose(layout):
    if layout == cutlass.RowMajor:
        return cutlass.ColumnMajor
    elif layout == cutlass.ColumnMajor:
        return cutlass.RowMajor
    elif layout == cutlass.ColumnMajorInterleaved32:
        return cutlass.RowMajorInterleaved32
    elif layout == cutlass.RowMajorInterleaved32:
        return cutlass.ColumnMajorInterleaved32


def getTensorRef(tensor: np.ndarray, problem_size: cutlass.gemm.GemmCoord, operand: str, layout: cutlass.layout):
    ptr = tensor.__array_interface__['data'][0]
    if operand == "a":
        tensor_coord = problem_size.mk()
    elif operand == "b":
        tensor_coord = problem_size.kn()
    elif operand in ["c", "d"]:
        tensor_coord = problem_size.mn()
    else:
        raise ValueError("unknonw operand: " + operand)

    if layout == cutlass.RowMajor:
        layout = cutlass.RowMajor.packed(tensor_coord)
        layout_tag = "RowMajor"
    elif layout == cutlass.ColumnMajor:
        layout = cutlass.ColumnMajor.packed(tensor_coord)
        layout_tag = "ColumnMajor"
    elif layout == cutlass.ColumnMajorInterleaved32:
        layout = cutlass.ColumnMajorInterleaved32.packed(tensor_coord)
        layout_tag = "ColumnMajorInterleaved32"
    elif layout == cutlass.RowMajorInterleaved32:
        layout = cutlass.RowMajorInterleaved32.packed(tensor_coord)
        layout_tag = "RowMajorInterleaved32"
    else:
        raise ValueError("unsupported layout")
    if tensor.dtype == np.float32:
        ref_name = "TensorRefF32" + layout_tag
    elif tensor.dtype == np.float64:
        ref_name = "TensorRefF64" + layout_tag
    elif tensor.dtype == np.float16:
        ref_name = "TensorRefF16" + layout_tag
    elif tensor.dtype == bfloat16:
        ref_name = "TensorRefBF16" + layout_tag
    elif tensor.dtype == np.int8:
        ref_name = "TensorRefS8" + layout_tag
    elif tensor.dtype == np.int32:
        ref_name = "TensorRefS32" + layout_tag
    else:
        raise ValueError("unsupported datatype %s" %
                         ShortDataTypeNames[tensor.dtype])

    return getattr(cutlass, ref_name)(ptr, layout)


def getTensorView(tensor: np.ndarray, problem_size: cutlass.gemm.GemmCoord, operand: str, layout: str):
    tensor_ref = getTensorRef(tensor, problem_size, operand, layout)

    if operand == "a":
        tensor_coord = problem_size.mk()
    elif operand == "b":
        tensor_coord = problem_size.kn()
    elif operand in ["c", "d"]:
        tensor_coord = problem_size.mn()
    else:
        raise ValueError("unknonw operand: " + operand)

    if layout == cutlass.RowMajor:
        layout_tag = "RowMajor"
    elif layout == cutlass.ColumnMajor:
        layout_tag = "ColumnMajor"
    elif layout == cutlass.ColumnMajorInterleaved32:
        layout_tag = "ColumnMajorInterleaved32"
    elif layout == cutlass.RowMajorInterleaved32:
        layout_tag = "RowMajorInterleaved32"
    else:
        raise ValueError("unsupported layout")
    if tensor.dtype == np.float32:
        ref_name = "TensorViewF32" + layout_tag
    elif tensor.dtype == np.float64:
        ref_name = "TensorViewF64" + layout_tag
    elif tensor.dtype == np.float16:
        ref_name = "TensorViewF16" + layout_tag
    elif tensor.dtype == bfloat16:
        ref_name = "TensorViewBF16" + layout_tag
    elif tensor.dtype == np.int32:
        ref_name = "TensorViewS32" + layout_tag
    elif tensor.dtype == np.int8:
        ref_name = "TensorViewS8" + layout_tag
    else:
        raise ValueError("unsupported datatype")

    return getattr(cutlass, ref_name)(tensor_ref, tensor_coord)


class GemmUniversalLauncher:
    def __init__(self, operation: 'GemmOperationUniversal', seed: int = 2080, interleaved=False,
                 verification=True, profiling=False, warmup_iterations=500, iterations=500, **kwargs) -> None:
        # create the reduction kernel
        self.reduction_operation: ReductionOperation = ReductionOperation(
            shape=cutlass.MatrixCoord(4, 32 * operation.C.alignment),
            C=operation.C, element_accumulator=operation.tile_description.math_instruction.element_accumulator,
            element_compute=operation.epilogue_functor.element_epilogue, epilogue_functor=operation.epilogue_functor,
            count=operation.C.alignment
        )

        self.math_operation = operation.tile_description.math_instruction.math_operation

        #: verify the output result
        self.verification = verification
        #: profile the kernel's runtime
        self.profiling = profiling

        self.timer = GpuTimer()

        self.warmup_iterations = warmup_iterations
        self.iterations = iterations

        if "sleep" in kwargs.keys():
            self.sleep_time = kwargs["sleep"]
        else:
            self.sleep_time = 0

        #
        # Compile the operator
        #

        pycutlass.compiler.add_module([operation, self.reduction_operation])

        self.operation = operation

        self.dtype_A = GemmUniversalLauncher.numpy_type(operation.A.element)
        self.dtype_B = GemmUniversalLauncher.numpy_type(operation.B.element)
        self.dtype_C = GemmUniversalLauncher.numpy_type(operation.C.element)
        self.dtype_D = GemmUniversalLauncher.numpy_type(operation.C.element)

        accumulator_size = DataTypeSize[operation.tile_description.math_instruction.element_accumulator]
        element_size = DataTypeSize[operation.A.element]

        if element_size == 1:
            self.scope_max = 1
            self.scope_min = 0
        elif element_size <= 8:
            self.scope_max = 1
            self.scope_min = -1
        elif element_size == 16:
            self.scope_max = 4
            self.scope_min = -4
        else:
            self.scope_max = 8
            self.scope_min = -8

        #: seed
        self.seed: int = seed

        #: whether the layout is interleaved
        self.interleaved = interleaved

        #: compute type
        self.compute_type = operation.epilogue_functor.element_epilogue
        self.accumulator_type = operation.tile_description.math_instruction.element_accumulator

    def print_problem_size(self, p, mode, batch_count):
        if mode == cutlass.gemm.Mode.Gemm:
            mode = "Gemm"
        elif mode == cutlass.gemm.Mode.GemmSplitKParallel:
            mode = "GemmSplitKParalel"
        problem_size = "problem: %d, %d, %d\n batch_count: %d\n mode: %s" % (
            p.m(), p.n(), p.k(), batch_count, mode)
        print(problem_size)

    @staticmethod
    def numpy_type(type):
        if type == cutlass.float64:
            return np.float64
        elif type == cutlass.float32:
            return np.float32
        elif type == cutlass.float16:
            return np.float16
        elif type == cutlass.bfloat16:
            return bfloat16
        elif type == cutlass.int32:
            return np.int32
        elif type == cutlass.int8:
            return np.int8
        else:
            raise ValueError("unsupported type: %s" % ShortDataTypeNames[type])

    def uniform_init(self, size, dtype):
        if dtype in [np.float32, np.float16, bfloat16, np.float64]:
            return np.ceil(
                np.random.uniform(
                    low=self.scope_min - 0.5, high=self.scope_max - 0.5,
                    size=size).astype(dtype)
            )
        else:
            return np.random.uniform(
                low=self.scope_min - 1, high=self.scope_max + 1,
                size=size).astype(dtype)

    def reorder_tensor_B(self, tensor_B, problem_size):
        reordered_tensor_B = np.empty_like(tensor_B)
        tensor_ref_B = getTensorRef(
            tensor_B, problem_size, "b", self.operation.B.layout)
        reordered_tensor_ref_B = getTensorRef(
            reordered_tensor_B, problem_size, "b", self.operation.B.layout)
        cutlass.gemm.host.reorder_column(
            tensor_ref_B, reordered_tensor_ref_B, problem_size)
        return reordered_tensor_B

    def host_reference(self, problem_size, tensor_A, tensor_B, tensor_C, alpha, beta):
        # TODO
        tensor_D_ref = np.ones_like(tensor_C)
        alpha = self.numpy_type(self.compute_type)(alpha)
        beta = self.numpy_type(self.compute_type)(beta)
        init_acc = 0

        alpha = self.compute_type(alpha).value()
        beta = self.compute_type(beta).value()
        init_acc = self.accumulator_type(init_acc).value()

        if self.operation.switched:
            tensor_ref_A = getTensorRef(
                tensor_A, problem_size, "a", transpose(self.operation.B.layout))
            tensor_ref_B = getTensorRef(
                tensor_B, problem_size, "b", transpose(self.operation.A.layout))
            tensor_ref_C = getTensorRef(
                tensor_C, problem_size, "c", transpose(self.operation.C.layout))
            tensor_ref_D_ref = getTensorRef(
                tensor_D_ref, problem_size, "d", transpose(self.operation.C.layout))
        else:
            tensor_ref_A = getTensorRef(
                tensor_A, problem_size, "a", self.operation.A.layout)
            tensor_ref_B = getTensorRef(
                tensor_B, problem_size, "b", self.operation.B.layout)
            tensor_ref_C = getTensorRef(
                tensor_C, problem_size, "c", self.operation.C.layout)
            tensor_ref_D_ref = getTensorRef(
                tensor_D_ref, problem_size, "d", self.operation.C.layout)

        if self.math_operation in [MathOperation.multiply_add_saturate]:
            cutlass.test.gemm.host.gemm_saturate(
                problem_size, alpha, tensor_ref_A, tensor_ref_B, beta, tensor_ref_C, tensor_ref_D_ref, init_acc)
        else:
            cutlass.test.gemm.host.gemm(problem_size, alpha, tensor_ref_A,
                                        tensor_ref_B, beta, tensor_ref_C, tensor_ref_D_ref, init_acc)

        return tensor_D_ref

    def equal(self, tensor_D, tensor_D_ref, problem_size):

        tensor_view_D = getTensorView(
            tensor_D, problem_size, "d", self.operation.C.layout)
        tensor_view_D_ref = getTensorView(
            tensor_D_ref, problem_size, "d", self.operation.C.layout)

        return cutlass.test.gemm.host.equals(tensor_view_D, tensor_view_D_ref)

    def bytes(self, problem_size, batch_count=1, alpha=1.0, beta=0.0):
        m = problem_size.m()
        n = problem_size.n()
        k = problem_size.k()

        bytes = \
            (DataTypeSize[self.operation.A.element] * m // 8) * k + \
            (DataTypeSize[self.operation.B.element] * n // 8) * k + \
            (DataTypeSize[self.operation.C.element] * m // 8) * n

        if beta != 0:
            bytes += (DataTypeSize[self.operation.C.element] * m // 8) * n

        bytes *= batch_count

        return bytes

    def flops(self, problem_size, batch_count=1):
        m = problem_size.m()
        n = problem_size.n()
        k = problem_size.k()

        flops_ = (m * n * k + m * n) * 2 * batch_count

        # TODO: complex
        return flops_

    def run_cutlass_profiler(self, mode, problem_size, batch_count=1, alpha=1.0, beta=0.0):

        cutlass_path = os.getenv('CUTLASS_PATH')
        assert cutlass_path is not None, "Environment variable 'CUTLASS_PATH' is not defined."

        values = {
            "profiler_path": cutlass_path + "/build/tools/profiler/cutlass_profiler",
            "kernel_name": self.operation.procedural_name(),
            "verification_providers": "device",
            "provider": "cutlass",
            "m": str(problem_size.m()),
            "n": str(problem_size.n()),
            "k": str(problem_size.k()),
            'split_k_slices': str(batch_count),
            'alpha': str(alpha),
            'beta': str(beta),
            'warmup': str(self.warmup_iterations),
            'profile': str(self.iterations)
        }

        cmd_template = \
            "${profiler_path} --kernels=${kernel_name} --verification-providers=${verification_providers}" \
            " --providers=${provider} --m=${m} --n=${n} --k=${k}"

        cmd = SubstituteTemplate(cmd_template, values)
        result = subprocess.getoutput(cmd)

        m = re.search(r"Runtime:\s+(?P<runtime>\d+.\d+)", result)
        runtime = float(m.group('runtime'))

        m = re.search(r"Bytes:\s+(?P<bytes>\d+)", result)
        bytes = int(m.group('bytes'))

        m = re.search(r"FLOPs:\s+(?P<flops>\d+)", result)
        flops = int(m.group('flops'))

        # check if the problem size matches
        assert bytes == self.bytes(problem_size, alpha, beta)
        assert flops == self.flops(problem_size)

        return runtime

    def run(self, mode, problem_size, batch_count=1, alpha=1.0, beta=0.0):

        assert get_allocated_size(
        ) == 0, "%d byte of pool memory is not released in previous run" % get_allocated_size()

        np.random.seed(self.seed)

        tensor_A = self.uniform_init(
            size=(problem_size.m() * problem_size.k(),), dtype=self.dtype_A)
        tensor_B = self.uniform_init(
            size=(problem_size.n() * problem_size.k(),), dtype=self.dtype_B)
        tensor_C = self.uniform_init(
            size=(problem_size.m() * problem_size.n(),), dtype=self.dtype_C)
        tensor_D = np.zeros(
            shape=(problem_size.m() * problem_size.n(),), dtype=self.dtype_D)

        #
        # Launch kernel
        #

        arguments = GemmArguments(
            operation=self.operation, problem_size=problem_size,
            A=tensor_A, B=tensor_B, C=tensor_C, D=tensor_D,
            output_op=self.operation.epilogue_type(alpha, beta),
            gemm_mode=mode, split_k_slices=batch_count
        )

        if mode == cutlass.gemm.Mode.GemmSplitKParallel:
            reduction_arguments = ReductionArguments(
                self.reduction_operation, problem_size=[
                    problem_size.m(), problem_size.n()],
                partitions=batch_count,
                workspace=arguments.ptr_D,
                destination=tensor_D,
                source=tensor_C,
                output_op=self.reduction_operation.epilogue_type(alpha, beta)
            )

        self.operation.run(arguments)

        if mode == cutlass.gemm.Mode.GemmSplitKParallel:
            self.reduction_operation.run(reduction_arguments)

        passed = True

        if self.verification:
            if mode == cutlass.gemm.Mode.GemmSplitKParallel:
                reduction_arguments.sync()
            else:
                arguments.sync()
            tensor_D_ref = self.host_reference(
                problem_size, tensor_A, tensor_B, tensor_C, alpha, beta)
            passed = self.equal(tensor_D, tensor_D_ref, problem_size)

            try:
                assert passed
            except AssertionError:
                self.print_problem_size(problem_size, mode, batch_count)

        if self.profiling:
            sleep(self.sleep_time)
            for _ in range(self.warmup_iterations):
                self.operation.run(arguments)
                if mode == cutlass.gemm.Mode.GemmSplitKParallel:
                    self.reduction_operation.run(reduction_arguments)

            self.timer.start()
            for _ in range(self.iterations):
                self.operation.run(arguments)
                if mode == cutlass.gemm.Mode.GemmSplitKParallel:
                    self.reduction_operation.run(reduction_arguments)
            self.timer.stop_and_wait()

            runtime = self.timer.duration(self.iterations)

        # free memory and clear buffers
        del arguments
        if mode == cutlass.gemm.Mode.GemmSplitKParallel:
            del reduction_arguments

        assert get_allocated_size(
        ) == 0, "%d byte of pool memory is not released after current run" % get_allocated_size()

        if self.profiling:
            return runtime
        return passed


def test_all_gemm(operation: 'GemmOperationUniversal', testcase="universal"):

    passed = True

    minimum_operand_element_size = min(
        DataTypeSize[operation.A.element], DataTypeSize[operation.B.element])
    opcode_class = operation.tile_description.math_instruction.opcode_class

    if opcode_class == cutlass.OpClass.Simt:
        alignment = 1
    else:
        alignment = 128 // minimum_operand_element_size

    # int8_t gemm alignment constrainst
    if opcode_class == cutlass.OpClass.Simt and operation.A.element == cutlass.int8 and operation.A.layout == cutlass.ColumnMajor:
        alignment_m = 4
    else:
        alignment_m = alignment

    if opcode_class == cutlass.OpClass.Simt and operation.B.element == cutlass.int8 and operation.A.layout == cutlass.RowMajor:
        alignment_n = 4
    else:
        alignment_n = alignment

    if opcode_class == cutlass.OpClass.Simt and operation.A.element == cutlass.int8 \
            and operation.B.element == cutlass.int8 \
            and (operation.A.layout == cutlass.RowMajor or operation.B.layout == cutlass.ColumnMajor):

        alignment_k = 4
    else:
        alignment_k = alignment

    threadblock_k = operation.tile_description.threadblock_shape[2]

    if testcase == "interleaved":
        if operation.A.layout in [cutlass.ColumnMajorInterleaved32, cutlass.RowMajorInterleaved32]:
            interleavedk = 32
        else:
            raise ValueError("unknonw layout")

    if testcase == "interleaved":
        modes = [cutlass.gemm.Mode.Gemm, ]
        problem_size_m = [interleavedk, 512+interleavedk]
        problem_size_n = [interleavedk, 512+interleavedk]
        problem_size_k = [interleavedk, threadblock_k *
                          operation.tile_description.stages + interleavedk]
        problem_alpha = [1.0]
        problem_beta = [0.0]
        batch_counts = [1, ]
    elif testcase == "multistage":
        modes = [cutlass.gemm.Mode.Gemm, ]
        problem_size_m = [16, 528]
        problem_size_n = [16, 528]
        problem_size_k = [threadblock_k, threadblock_k * operation.tile_description.stages +
                          operation.tile_description.math_instruction.instruction_shape[2]]
        problem_alpha = [1.0]
        problem_beta = [0.0]
        batch_counts = [1, ]
    else:  # universal
        modes = [cutlass.gemm.Mode.Gemm, cutlass.gemm.Mode.GemmSplitKParallel]
        problem_size_m = [alignment_m, 512 - 3 * alignment_m]
        problem_size_n = [alignment_n, 512 - 2 * alignment_n]
        problem_size_k = [
            alignment_k,
            threadblock_k * operation.tile_description.stages - alignment_k,
            threadblock_k * operation.tile_description.stages * 3 - alignment_k]
        batch_counts = [1, 2, 3, 5, 7]
        problem_alpha = [1.0]
        problem_beta = [2.0]

    testbed = GemmUniversalLauncher(
        operation, interleaved=(testcase == "interleaved"))

    for mode in modes:
        for m in problem_size_m:
            for n in problem_size_n:
                for k in problem_size_k:
                    for batch_count in batch_counts:
                        for alpha in problem_alpha:
                            for beta in problem_beta:
                                # skip very small K problems
                                if testcase == "universal":
                                    if (k // batch_count < 2 * threadblock_k):
                                        continue

                                problem_size = cutlass.gemm.GemmCoord(m, n, k)

                                passed = testbed.run(
                                    mode, problem_size, batch_count, alpha, beta)

                                err, = cudart.cudaDeviceSynchronize()
                                if err != cuda.CUresult.CUDA_SUCCESS:
                                    raise RuntimeError(
                                        "CUDA Error %s" % str(err))

                                if not passed:
                                    return False

    return passed



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

from cuda import cuda
from cuda import cudart


class GpuTimer:
    def __init__(self) -> None:
        self.events = [
            cuda.cuEventCreate(cuda.CUevent_flags.CU_EVENT_DEFAULT)[1],
            cuda.cuEventCreate(cuda.CUevent_flags.CU_EVENT_DEFAULT)[1]
        ]

    def start(self, stream=cuda.CUstream(0)):
        err, = cuda.cuEventRecord(self.events[0], stream)
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError("CUDA Error %s" % str(err))

    def stop(self, stream=cuda.CUstream(0)):
        err, = cuda.cuEventRecord(self.events[1], stream)
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError("CUDA Error %s" % str(err))
        pass

    def stop_and_wait(self, stream=cuda.CUstream(0)):
        self.stop(stream)
        if stream:
            err, = cuda.cuStreamSynchronize(stream)
            if err != cuda.CUresult.CUDA_SUCCESS:
                raise RuntimeError("CUDA Error %s" % str(err))
        else:
            err, = cudart.cudaDeviceSynchronize()
            if err != cuda.CUresult.CUDA_SUCCESS:
                raise RuntimeError("CUDA Error %s" % str(err))

    def duration(self, iterations=1):
        err, duration = cuda.cuEventElapsedTime(self.events[0], self.events[1])
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError("CUDA Error %s" % str(err))
        return duration / float(iterations)



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################
## Test case for Pytorch
import pycutlass
import unittest
from pycutlass import *
import torch
import cupy as cp


class Test_Frontend(unittest.TestCase):
    def setUp(self) -> None:
        #
        # define the cutlass operator
        #
        math_inst = MathInstruction(
            [1, 1, 1], cutlass.float32, cutlass.float32, cutlass.float32,
            cutlass.OpClass.Simt, MathOperation.multiply_add
        )

        tile_description = TileDescription(
            [128, 128, 8], 4, [2, 4, 1],
            math_inst
        )

        A = TensorDescription(
            cutlass.float32, cutlass.RowMajor, 1
        )

        B = TensorDescription(
            cutlass.float32, cutlass.RowMajor, 1
        )

        C = TensorDescription(
            cutlass.float32, cutlass.RowMajor, 1
        )

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, cutlass.float32)

        self.operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, 
            swizzling_functor=cutlass.IdentitySwizzle1
        )

        pycutlass.compiler.add_module([self.operation,])


    def test_torch_frontend(self):
        problem_size = cutlass.gemm.GemmCoord(512, 256, 128)

        tensor_A = torch.ceil(torch.empty(size=(problem_size.m(), problem_size.k()), dtype=torch.float32, device="cuda").uniform_(-8.5, 7.5))
        tensor_B = torch.ceil(torch.empty(size=(problem_size.k(), problem_size.n()), dtype=torch.float32, device="cuda").uniform_(-8.5, 7.5))
        tensor_C = torch.ceil(torch.empty(size=(problem_size.m(), problem_size.n()), dtype=torch.float32, device="cuda").uniform_(-8.5, 7.5))
        tensor_D = torch.empty_like(tensor_C)
        

        alpha = 1.0
        beta = 0.0

        arguments = GemmArguments(
            operation=self.operation, problem_size=problem_size,
            A=tensor_A, B=tensor_B, C=tensor_C, D=tensor_D,
            output_op=self.operation.epilogue_type(alpha, beta),
            gemm_mode=cutlass.gemm.Mode.Gemm, split_k_splices=1
        )

        self.operation.run(arguments)

        arguments.sync()

        tensor_D_ref = alpha * tensor_A @ tensor_B + beta * tensor_C

        self.assertTrue(torch.equal(tensor_D, tensor_D_ref))
    
    def test_cupy_frontend(self):
        cp.cuda.set_allocator(rmm.rmm_cupy_allocator)

        problem_size = cutlass.gemm.GemmCoord(512, 256, 128)

        tensor_A = cp.ceil(cp.random.uniform(low=-8.5, high=7.5, size=(problem_size.m(), problem_size.k()), dtype=cp.float32))
        tensor_B = cp.ceil(cp.random.uniform(low=-8.5, high=7.5, size=(problem_size.k(), problem_size.n()), dtype=cp.float32))
        tensor_C = cp.ceil(cp.random.uniform(low=-8.5, high=7.5, size=(problem_size.m(), problem_size.n()), dtype=cp.float32))
        tensor_D = cp.ones_like(tensor_C)

        alpha = 1.0
        beta = 1.0

        tensor_D_ref = alpha * tensor_A @ tensor_B + beta * tensor_C

        arguments = GemmArguments(
            operation=self.operation, problem_size=problem_size,
            A=tensor_A, B=tensor_B, C=tensor_C, D=tensor_D,
            output_op=self.operation.epilogue_type(alpha, beta),
            gemm_mode=cutlass.gemm.Mode.Gemm, split_k_splices=1
        )

        self.operation.run(arguments)

        arguments.sync()

        self.assertTrue(cp.array_equal(tensor_D, tensor_D_ref))



if __name__ == '__main__':
    pycutlass.get_memory_pool(2**32, 2**32)
    unittest.main()



################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################
from typeguard import typechecked
from cuda import cuda
from typing import Union
import numpy as np

from typeguard import typechecked

from pycutlass import *


# @typechecked
class Conv2dArguments(ArgumentBase):
    """
    Argument wrapper for Conv2d. It encodes problem information and 
    user-provide tensors into the kernel's argument.

    :param operation: the Conv2d operation to take the argument
    :type operation: :class:`pycutlass.Conv2dOperation`

    :param problem_size: the Conv2d problem size
    :type problem_size: :class:`cutlass.conv.Conv2dProblemSize`

    :param A: tensor A
    :type A: cuda.CUdeviceptr | numpy.ndarray | torch.Tensor | cupy.ndarray

    :param B: tensor B
    :type B: cuda.CUdeviceptr | numpy.ndarray | torch.Tensor | cupy.ndarray

    :param C: tensor C
    :type C: cuda.CUdeviceptr | numpy.ndarray | torch.Tensor | cupy.ndarray

    :param D: tensor D
    :type D: cuda.CUdeviceptr | numpy.ndarray | torch.Tensor | cupy.ndarray

    :param split_k_mode: conv2d split K mode, defaults to 
    cutlass.conv.SplitKMode.Serial
    :type split_k_mode: cutlass.conv.SplitKMode, optional

    :param output_op: output operator, optional
    :type output_op: :class:`pycutlass.LinearCombinationFunctorArguments`

    """

    def __init__(self, operation: 'Conv2dOperation',
                 problem_size: 'cutlass.conv.Conv2dProblemSize',
                 A: 'Union[cuda.CUdeviceptr, np.ndarray, torch.Tensor]',
                 B: 'Union[cuda.CUdeviceptr, np.ndarray, torch.Tensor]',
                 C: 'Union[cuda.CUdeviceptr, np.ndarray, torch.Tensor]',
                 D: 'Union[cuda.CUdeviceptr, np.ndarray, torch.Tensor]',
                 split_k_mode: 'cutlass.conv.SplitKMode'
                    = cutlass.conv.SplitKMode.Serial, **kwargs) -> None:

        self.operation = operation
        #: convolution kind
        self.conv_kind: cutlass.conv.Operator = operation.conv_kind
        self.layout_A: cutlass.layout = operation.A.layout
        self.layout_B: cutlass.layout = operation.B.layout
        self.layout_C: cutlass.layout = operation.C.layout

        self.element_A = operation.A.element
        self.element_B = operation.B.element
        self.element_C = operation.C.element

        if self.layout_C == cutlass.TensorNC32HW32:
            B = self.reorder_tensor_B(B, problem_size)

        super().__init__(A, B, C, D, **kwargs)
        # preprocessing output ops
        
        if 'output_op' in kwargs.keys() and \
            split_k_mode != cutlass.conv.SplitKMode.Parallel:
            self.output_op = kwargs['output_op']
        else:
            self.output_op = self.operation.epilogue_type(1.0, 0.0)

        if "split_k_slices" in kwargs.keys():
            self.split_k_mode = split_k_mode
            self.split_k_slices = kwargs["split_k_slices"]
        else:
            self.split_k_mode = cutlass.conv.SplitKMode.Serial
            self.split_k_slices = 1

        #: problem_size
        self.problem_size: cutlass.conv.Conv2dProblemSize = problem_size
        self.problem_size.split_k_slices = self.split_k_slices

        if hasattr(self, "tensor_c_numel"):
            c_coord = cutlass.conv.implicit_gemm_tensor_c_extent(
                self.conv_kind, problem_size)
            if (self.tensor_c_numel == c_coord.at(3) and 
                self.tensor_c_numel < c_coord.size()):
                self.bias = True

        #
        # initialize the argument
        #
        self.initialize()

    # @typechecked
    def reorder_tensor_B(self, tensor_B: 'np.ndarray', 
            problem_size: 'cutlass.conv.Conv2dProblemSize'):
        """
        Reorder tensor_B for interleaved layout

        :param tensor_B: input tensor B
        :type tensor_B: numpy.ndarray
        :param problem_size: Conv2d problem size
        :type problem_size: :class:`cutlass.conv.Conv2dProblemSize`

        :return: reordered tensor B
        :rtype: numpy.ndarray
        """
        reordered_tensor_B = np.empty_like(tensor_B)
        tensor_ref_B = self.get_tensor_ref(
            tensor_B, self.element_B, self.layout_B, problem_size, "b")
        reordered_tensor_ref_B = self.get_tensor_ref(
            reordered_tensor_B, self.element_B, 
            self.layout_B, problem_size, "b")
        cutlass.conv.host.reorder_convK(
            reordered_tensor_ref_B, tensor_ref_B, self.conv_kind, problem_size)

        return reordered_tensor_B

    def get_tensor_ref(
        self, tensor, dtype, tensor_layout, problem_size, operand):
        if operand == "a":
            tensor_coord = cutlass.conv.implicit_gemm_tensor_a_extent(
                self.conv_kind, problem_size)
        elif operand == "b":
            tensor_coord = cutlass.conv.implicit_gemm_tensor_b_extent(
                self.conv_kind, problem_size)
        elif operand in ["c", "d"]:
            tensor_coord = cutlass.conv.implicit_gemm_tensor_c_extent(
                self.conv_kind, problem_size)
        else:
            raise ValueError("unknown operand: " + operand)
        # Zero stride trick
        if operand == "c" and self.bias:
            tensor_coord = cutlass.Tensor4DCoord(0, 0, 0, 0)

        layout = tensor_layout.packed(tensor_coord)

        return TensorRef(tensor, dtype, layout).tensor_ref

    def get_arguments(self, semaphore):
        ref_A = TensorRef_(self.get_tensor_ref(
            self.ptr_A, self.element_A, self.layout_A, self.problem_size, "a"))
        ref_B = TensorRef_(self.get_tensor_ref(
            self.ptr_B, self.element_B, self.layout_B, self.problem_size, "b"))
        ref_C = TensorRef_(self.get_tensor_ref(
            self.ptr_C, self.element_C, self.layout_C, self.problem_size, "c"))
        ref_D = TensorRef_(self.get_tensor_ref(
            self.ptr_D, self.element_C, self.layout_C, self.problem_size, "d"))

        self.c_arguments = self.operation.argument_type(
            Conv2DProblemSize(self.problem_size),
            ref_A, ref_B, ref_C, ref_D, self.output_op, self.split_k_mode
        )

        self.semaphore = semaphore

    def initialize(self):
        """
        Initialize the kernel arguments handling following stuffs
        1. get kernel launch configuration including grid, cta size, 
           and dynamic shared memory capacity
        2. allocate and initialize device workspace
        3. get kernel params as bytearray for NVRTC input
        """
        # get launch configuration
        self.launch_config = self.operation.rt_module.plan(self)

        # allocate and initialize device workspace
        device_workspace_size = \
            self.operation.rt_module.get_device_workspace_size(self)

        if device_workspace_size > 0:
            self.workspace_buffer = device_mem_alloc(device_workspace_size)
            workspace_ptr = self.workspace_buffer.ptr
            err, = cuda.cuMemsetD32(
                workspace_ptr, 0, device_workspace_size // 4)
        else:
            workspace_ptr = None

        # get kernel params as bytearray
        semaphore = 0
        if workspace_ptr is not None and \
            self.split_k_mode == cutlass.conv.SplitKMode.Parallel:
            self.ptr_D = workspace_ptr
        elif workspace_ptr is not None and \
            self.split_k_mode == cutlass.conv.SplitKMode.Serial:
            semaphore = workspace_ptr

        self.get_arguments(semaphore)

        params_ = self.operation.rt_module.get_args(ctypes.byref(
            self.c_arguments), ctypes.c_void_p(int(self.semaphore)))
        self.host_workspace = bytearray(params_.contents)
        self.device_workspace = None

    def sync(self):
        """
        Synchronize the arguments. If the input tensor is in host, 
        copy it from device to host.
        """
        return super().sync()


# @typechecked
class Conv2dRT(ExecutableOperation):
    """
    Conv2dRT manages the CUTLASS runtime components
    """
    KernelTemplate = r'''
extern "C"
__global__ void
${operation_name}(${operation_name}${operation_suffix}::Params params) {

  // Dynamic shared memory base pointer
  extern __shared__ int SharedStorageBase[];

  // Declare pointer to dynamic shared memory.
  ${operation_name}${operation_suffix}::SharedStorage *shared_storage =
      reinterpret_cast<${operation_name}${operation_suffix}::SharedStorage *>(SharedStorageBase);

  ${operation_name}${operation_suffix} op;

  op(params, *shared_storage);
}
    '''

    HostTemplate = r'''
extern "C" {
  // Get the size of params in bytes
  int ${operation_name}_get_param_size(){
    return sizeof(${operation_name}${operation_suffix}::Params);
  }

  // Get the size of dynamic shared memory in bytes
  int ${operation_name}_shared_memory_size() {
    return int(sizeof(${operation_name}${operation_suffix}::SharedStorage));
  }

  // Get the params as byte array
  char* ${operation_name}_get_params(${operation_name}${operation_suffix}::Arguments* arguments, int *semaphore=nullptr){
    typename ${operation_name}${operation_suffix}::Params* params;
    params = new ${operation_name}${operation_suffix}::Params(*arguments, semaphore);

    char *bytes = ((char*)(params));
    char *output = new char[sizeof(${operation_name}${operation_suffix}::Params)];
    for (unsigned int i = 0; i < sizeof(${operation_name}${operation_suffix}::Params); i ++)
        output[i] = bytes[i];

    return output;
  }
}

    '''

    def __init__(self, operation: 'Conv2dOperation'):
        super().__init__(operation)
        self.argument_type, self.epilogue_type = get_conv2d_arguments(operation.epilogue_functor)
        self.argtype = [ctypes.POINTER(self.argument_type), ctypes.c_void_p]
        self.conv_kind = operation.conv_kind

        self.operation: Conv2dOperation = operation

        self.emitter = EmitConv2dInstance('_type')

        self.threads: int = operation.tile_description.num_threads

        self.swizzle_functor = operation.swizzling_functor

    def emit(self):
        return self.emitter.emit(self.operation)

    # @typechecked
    def get_device_workspace_size(self, arguments: Conv2dArguments):
        workspace_bytes = 0

        launch_config = arguments.launch_config

        self.conv_kind = self.operation.conv_kind

        if arguments.split_k_mode == cutlass.conv.SplitKMode.Parallel:
            problem_size = arguments.problem_size
            workspace_bytes = DataTypeSize[self.operation.C.element] \
            * launch_config.grid[2] * cutlass.conv.implicit_gemm_tensor_c_size(
                self.conv_kind, problem_size
            ) // 8
        elif arguments.split_k_mode == cutlass.conv.SplitKMode.Serial and \
            arguments.split_k_slices > 1:
            workspace_bytes = launch_config.grid[0] * launch_config.grid[1] * 4

        return workspace_bytes

    # @typechecked
    def plan(self, arguments: Conv2dArguments):
        tile_size = cutlass.gemm.GemmCoord(
            self.operation.tile_description.threadblock_shape[0],
            self.operation.tile_description.threadblock_shape[1],
            self.operation.tile_description.threadblock_shape[2]
        )

        grid = self.swizzle_functor.get_grid_shape(
            self.swizzle_functor.get_tiled_shape(
                self.conv_kind, arguments.problem_size, 
                tile_size, arguments.split_k_slices
            )
        )
        return LaunchConfiguration(
            [grid.x, grid.y, grid.z], [self.threads, 1, 1], 
            self.shared_memory_capacity)

    def initialize(self):
        err, = cuda.cuFuncSetAttribute(
            self.kernel,
            attrib=cuda.CUfunction_attribute.CU_FUNC_ATTRIBUTE_MAX_DYNAMIC_SHARED_SIZE_BYTES,
            value=self.shared_memory_capacity)
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError('Cuda Error: {}'.format(err))

#


class Conv2dOperation:
    """
    CUTLASS Conv2d operation description.

    :param conv_kind: convolution operator
    :type conv_kind: :class:`cutlass.conv.Operator`

    :param iterator_algorithm: Selects among several implementation 
    variants trading off performance with simplicity
    :type iterator_algorithm: :class:`cutlass.conv.IteratorAlgorithm`

    :param arch: GPU compute capability (sm_xx)
    :type arch: int

    :param tile_description: tile description
    :type tile_description: :class:`pycutlass.TileDescription`

    :param A: tensor A description
    :type A: :class:`pycutlass.TensorDescription`

    :param B: tensor B description
    :type B: :class:`pycutlass.TensorDescription`

    :param C: tensor C description
    :type C: :class:`pycutlass.TensorDescription`

    :param D: tensor D description
    :type D: :class:`pycutlass.TensorDescription`

    :param element_epilogue: element type for computation in epilogue \
    :type element_epilogue: cutlass.int8 | cutlass.int32 | cutlass.float16 | \
    cutlass.bfloat16 | cutlass.float32 | cutlass.float64

    :param stride_support: distinguish among partial specializations that \
    accelerate certain problems where convolution stride is unit \
    :type stride_support: :class:`cutlass.conv.StrideSupport`

    :param epilogue_functor: convolution epilogue functor
    :type epilogue_functor: :class:`EpilogueFunctor`

    :param swizzling_functor: threadblock swizzling functor
    """
    #

    def __init__(self,
                 conv_kind: cutlass.conv.Operator,
                 iterator_algorithm: cutlass.conv.IteratorAlgorithm,
                 arch: int, tile_description: TileDescription,
                 A: TensorDescription, B: TensorDescription, C: TensorDescription,
                 stride_support, epilogue_functor,
                 swizzling_functor=cutlass.IdentitySwizzle1):

        self.operation_kind: OperationKind = OperationKind.Conv2d
        self.arch: int = arch
        self.tile_description: TileDescription = tile_description
        self.conv_kind = conv_kind
        self.A: TensorDescription = A
        self.B: TensorDescription = B
        self.C: TensorDescription = C
        self.epilogue_functor = epilogue_functor
        self.iterator_algorithm = iterator_algorithm
        self.stride_support = stride_support
        self.swizzling_functor = swizzling_functor()

        self.rt_module: Conv2dRT = Conv2dRT(self)
        self.argument_type = self.rt_module.argument_type
        self.epilogue_type = self.rt_module.epilogue_type

    def run(self, arguments: Conv2dArguments) -> cuda.CUresult:
        """
        Launch the cuda kernel with input arguments

        :param arguments: conv2d arguments
        :type arguments: :class:`pycutlass.Conv2dArguments`
        """

        # launch the kernel
        err = self.rt_module.run(
            arguments.host_workspace,
            arguments.device_workspace,
            arguments.launch_config)

        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError('CUDA Error %s' % str(err))

        return err

    #
    # Get function name
    #

    def procedural_name(self):
        ''' The full procedural name indicates architecture, extended name, tile size, and layout. '''
        return self.configuration_name()
    #

    def configuration_name(self):
        ''' The full procedural name indicates architecture, extended name, tile size, and layout. '''

        opcode_class_name = OpcodeClassNames[self.tile_description.math_instruction.opcode_class]

        threadblock = "%dx%d_%dx%d" % (
            self.tile_description.threadblock_shape[0],
            self.tile_description.threadblock_shape[1],
            self.tile_description.threadblock_shape[2],
            self.tile_description.stages
        )

        if self.stride_support == StrideSupport.Unity:
            configuration_name = "cutlass_${opcode_class}_${extended_name}_${threadblock}_${layout}_unity_stride_align${alignment}"
        else:
            configuration_name = "cutlass_${opcode_class}_${extended_name}_${threadblock}_${layout}_align${alignment}"

        return SubstituteTemplate(
            configuration_name,
            {
                'opcode_class': opcode_class_name,
                'extended_name': self.extended_name(),
                'threadblock': threadblock,
                'layout': self.layout_name(),
                'alignment': "%d" % self.A.alignment,
            }
        )

    #
    def extended_name(self):
        ''' Append data types if they differ from compute type. '''
        if self.C.element != self.tile_description.math_instruction.element_accumulator and \
                self.A.element != self.tile_description.math_instruction.element_accumulator:
            extended_name = "${element_c}_${core_name}_${element_a}"
        elif self.C.element == self.tile_description.math_instruction.element_accumulator and  \
                self.A.element != self.tile_description.math_instruction.element_accumulator:
            extended_name = "${core_name}_${element_a}"
        else:
            extended_name = "${core_name}"

        extended_name = SubstituteTemplate(extended_name, {
            'element_a': DataTypeNames[self.A.element],
            'element_c': DataTypeNames[self.C.element],
            'core_name': self.core_name()
        })

        return extended_name

    #
    def layout_name(self):
        return "%s" % (ShortLayoutTypeNames[self.A.layout])

    #
    def core_name(self):
        ''' The basic operation kind is prefixed with a letter indicating the accumulation type. '''

        intermediate_type = ''

        if self.tile_description.math_instruction.opcode_class == cutlass.OpClass.TensorOp:
            inst_shape = "%d%d%d" % tuple(
                self.tile_description.math_instruction.instruction_shape)
            if self.tile_description.math_instruction.element_a != self.A.element and \
                    self.tile_description.math_instruction.element_a != self.accumulator_type():
                intermediate_type = DataTypeNames[self.tile_description.math_instruction.element_a]
        else:
            inst_shape = ''

        return "%s%s%s%s_%s" % (ShortDataTypeNames[self.accumulator_type()],
                                inst_shape, intermediate_type, ConvKindNames[self.conv_kind], IteratorAlgorithmNames[self.iterator_algorithm])

    #
    def is_complex(self):
        complex_operators = [
            MathOperation.multiply_add_complex,
            MathOperation.multiply_add_complex_gaussian
        ]
        return self.tile_description.math_instruction.math_operation in complex_operators

    #
    def accumulator_type(self):
        accum = self.tile_description.math_instruction.element_accumulator

        if self.is_complex():
            return get_complex_from_real(accum)

        return accum


###################################################################################################
#
# Emits single instances of a CUTLASS device-wide operator
#
###################################################################################################

class EmitConv2dInstance:
    def __init__(self, operation_suffix=''):
        self.operation_suffix = operation_suffix
        self.includes = [
            "cutlass/cutlass.h",
            "cutlass/conv/kernel/default_conv2d_fprop.h",
            "cutlass/conv/kernel/default_conv2d_dgrad.h",
            "cutlass/conv/kernel/default_conv2d_wgrad.h"
        ]
        self.template = """
// Conv2d${conv_kind_name} ${iterator_algorithm_name} kernel instance "${operation_name}"
using ${operation_name}_base = 
typename cutlass::conv::kernel::DefaultConv2d${conv_kind_name}<
  ${element_a}, 
  ${layout_a},
  ${element_b}, 
  ${layout_b},
  ${element_c}, 
  ${layout_c},
  ${element_accumulator},
  ${opcode_class},
  ${arch},
  cutlass::gemm::GemmShape<${threadblock_shape_m}, ${threadblock_shape_n}, ${threadblock_shape_k}>,
  cutlass::gemm::GemmShape<${warp_shape_m}, ${warp_shape_n}, ${warp_shape_k} >,
  cutlass::gemm::GemmShape<${instruction_shape_m}, ${instruction_shape_n}, ${instruction_shape_k}>,
  ${epilogue_functor},
  ${swizzling_functor}, // cutlass::gemm::threadblock::GemmSplitKIdentityThreadblockSwizzle<>,
  ${stages},
  ${math_operator},
  ${iterator_algorithm},
  ${stride_support},
  ${align_a},
  ${align_b}
>::Kernel;

struct ${operation_name}${operation_suffix}:
  public ${operation_name}_base { };

"""

    def emit(self, operation):

        warp_shape = [int(operation.tile_description.threadblock_shape[idx] /
                          operation.tile_description.warp_count[idx]) for idx in range(3)]

        epilogue_vector_length = int(min(
            operation.C.alignment * DataTypeSize[operation.C.element], 128) / DataTypeSize[operation.C.element])

        values = {
            'operation_name': operation.procedural_name(),
            'operation_suffix': self.operation_suffix,
            'conv_kind': ConvKindTag[operation.conv_kind],
            'conv_kind_name': ConvKindNames[operation.conv_kind].capitalize(),
            'element_a': DataTypeTag[operation.A.element],
            'layout_a': LayoutTag[operation.A.layout],
            'element_b': DataTypeTag[operation.B.element],
            'layout_b': LayoutTag[operation.B.layout],
            'element_c': DataTypeTag[operation.C.element],
            'layout_c': LayoutTag[operation.C.layout],
            'element_accumulator': DataTypeTag[operation.accumulator_type()],
            'opcode_class': OpcodeClassTag[operation.tile_description.math_instruction.opcode_class],
            'arch': "cutlass::arch::Sm%d" % operation.arch,
            'threadblock_shape_m': str(operation.tile_description.threadblock_shape[0]),
            'threadblock_shape_n': str(operation.tile_description.threadblock_shape[1]),
            'threadblock_shape_k': str(operation.tile_description.threadblock_shape[2]),
            'warp_shape_m': str(warp_shape[0]),
            'warp_shape_n': str(warp_shape[1]),
            'warp_shape_k': str(warp_shape[2]),
            'instruction_shape_m': str(operation.tile_description.math_instruction.instruction_shape[0]),
            'instruction_shape_n': str(operation.tile_description.math_instruction.instruction_shape[1]),
            'instruction_shape_k': str(operation.tile_description.math_instruction.instruction_shape[2]),
            'epilogue_vector_length': str(epilogue_vector_length),
            'epilogue_functor': operation.epilogue_functor.emit(),
            'swizzling_functor': operation.swizzling_functor.tag(),
            'stages': str(operation.tile_description.stages),
            'iterator_algorithm': IteratorAlgorithmTag[operation.iterator_algorithm],
            'iterator_algorithm_name': IteratorAlgorithmNames[operation.iterator_algorithm].capitalize(),
            'stride_support': StrideSupportTag[operation.stride_support],
            'math_operator': 'cutlass::arch::OpMultiplyAddComplex' if operation.is_complex() else
            MathOperationTag[operation.tile_description.math_instruction.math_operation],
            'align_a': str(operation.A.alignment),
            'align_b': str(operation.B.alignment),
        }

        return SubstituteTemplate(self.template, values)
